package com.example;

import akka.actor.typed.ActorRef;
import akka.actor.typed.Behavior;
import akka.actor.typed.javadsl.*;

import java.time.Duration;
import java.util.Objects;

public class Actor extends AbstractBehavior<Actor.Message> {


    //Timer, um bei den Berechnungen eine Sekunde lang zu warten
    TimerScheduler<Message> timer;

    Expression expression;

    //Zeichen für die Operation (Add, Sub, Mul) welche benutzt wird
    String operationString;

    //Referenz auf den initialen Actor
    ActorRef<PrintAndEvaluate.Message> initial;

    //Referenz auf den (potenziellen) OberSubActor. Es ist immer entweder initial oder oberActor null.
    ActorRef<Actor.Message> oberActor;

    //Linke Seite: true, Rechte Seite: false um zu wissen, ob dieser Actor ein linkes oder rechtes Kind ist.
    boolean side;

    //Zurückgegebener String & Integer von der linken Seite
    String leftString;
    int leftInt;

    //Zurückgegebener String & Integer von der rechten Seite
    String rightString;
    int rightInt;

    public interface Message {}

    //Message kommt von dem initialen Actor
    public record PrintMessage(ActorRef<PrintAndEvaluate.Message> sender, Expression expression) implements Message{}
    public record SubPrintMessage(ActorRef<Actor.Message> sender, Expression expression, boolean side) implements Message{}

    //Message kommt von einem Actor
    public record EvaluateMessage(ActorRef<com.example.PrintAndEvaluate.Message> sender, Expression expression) implements Message{}
    public record SubEvaluateMessage(ActorRef<Actor.Message> sender, Expression expression, boolean side) implements Message{}

    //Antwort von dem linken Kind
    public record LeftStringResponse(String string) implements Message{}
    public record LeftIntResponse(int wert) implements Message{}

    //Antwort von dem rechten Kind
    public record RightStringResponse(String string) implements Message{}
    public record RightIntResponse(int wert) implements Message{}

    //Nachricht an sich selbst, um nach einer Sekunde Wartezeit das Ergebnis zu berechnen
    public record Compute() implements Message{}

    private Actor(ActorContext<Actor.Message> context, TimerScheduler<Message> timer){
        super(context);
        this.timer = timer;
    }

    public static Behavior<Actor.Message> create(){
        return Behaviors.setup(context -> Behaviors.withTimers(timers ->  new Actor(context,timers)));
    }

    @Override
    public Receive<Actor.Message> createReceive(){
        return newReceiveBuilder()
                .onMessage(PrintMessage.class, this::onPrintMessage)
                .onMessage(SubPrintMessage.class, this::onSubPrintMessage)
                .onMessage(LeftStringResponse.class, this::onLeftStringResponse)
                .onMessage(RightStringResponse.class, this::onRightStringResponse)
                .onMessage(EvaluateMessage.class, this::onEvaluateMessage)
                .onMessage(SubEvaluateMessage.class, this::onSubEvaluateMessage)
                .onMessage(LeftIntResponse.class, this::onLeftIntResponse)
                .onMessage(RightIntResponse.class, this::onRightIntResponse)
                .onMessage(Compute.class, this::onComputeMessage)
                .build();
    }

    //Die Nachricht von dem initialen Actor wird verarbeitet und, falls nötig (kein Val), weitere Actors erstellt.
    //Dabei wird seine Referenz in @initial gespeichert und diese ist für den Formatierungsvorgang
    private Behavior<Message> onPrintMessage(PrintMessage message){
        this.initial = message.sender;
        this.expression = message.expression;
        //Wenn initial nur Val übergeben wird → einfach ausgeben ansonsten UnterActors erstellen
        if (Objects.requireNonNull(expression) instanceof Expression.Add add) {
            this.operationString = "+";
            ActorRef<Actor.Message> LeftSubActor = getContext().spawnAnonymous(Actor.create());
            LeftSubActor.tell(new Actor.SubPrintMessage(getContext().getSelf(),add.left(), true));
            ActorRef<Actor.Message> RightSubActor = getContext().spawnAnonymous(Actor.create());
            RightSubActor.tell(new Actor.SubPrintMessage(getContext().getSelf(),add.right(), false));
        }
        else if (expression instanceof Expression.Sub sub) {
            this.operationString = "-";
            ActorRef<Actor.Message> LeftSubActor = getContext().spawnAnonymous(Actor.create());
            LeftSubActor.tell(new Actor.SubPrintMessage(getContext().getSelf(),sub.left(), true));
            ActorRef<Actor.Message> RightSubActor = getContext().spawnAnonymous(Actor.create());
            RightSubActor.tell(new Actor.SubPrintMessage(getContext().getSelf(),sub.right(), false));
        }
        else if (expression instanceof Expression.Mul mul) {
            this.operationString = "*";
            ActorRef<Actor.Message> LeftSubActor = getContext().spawnAnonymous(Actor.create());
            LeftSubActor.tell(new Actor.SubPrintMessage(getContext().getSelf(),mul.left(), true));
            ActorRef<Actor.Message> RightSubActor = getContext().spawnAnonymous(Actor.create());
            RightSubActor.tell(new Actor.SubPrintMessage(getContext().getSelf(),mul.right(), false));
        } else if (expression instanceof Expression.Val val) {
            initial.tell(new PrintAndEvaluate.StringResponse(String.valueOf(val.inner())));
        }

        return this;
    }

    //Nachricht von dem Actor, von welchem dieser Actor erstellt wurde, wird verarbeitet und,
    //falls nötig (kein Val), weitere SubActors erstellt.
    //Dabei wird seine Referenz in oberActor gespeichert und diese Funktion ist für den Formatierungsvorgang
    private Behavior<Message> onSubPrintMessage(SubPrintMessage message){
        this.oberActor = message.sender;
        this.expression = message.expression;
        this.side = message.side;
        //Wenn initial nur Val übergeben wird → einfach ausgeben ansonsten UnterActors erstellen
        if (Objects.requireNonNull(expression) instanceof Expression.Add add) {
            this.operationString = "+";
            ActorRef<Message> subActorLeft = getContext().spawnAnonymous(Actor.create());
            subActorLeft.tell(new SubPrintMessage(getContext().getSelf(),add.left(), true));
            ActorRef<Message> subActorRight = getContext().spawnAnonymous(Actor.create());
            subActorRight.tell(new SubPrintMessage(getContext().getSelf(),add.right(), false));
        }
        else if (expression instanceof Expression.Sub sub) {
            this.operationString = "-";
            ActorRef<Message> subActorLeft = getContext().spawnAnonymous(Actor.create());
            subActorLeft.tell(new SubPrintMessage(getContext().getSelf(),sub.left(), true));
            ActorRef<Message> subActorRight = getContext().spawnAnonymous(Actor.create());
            subActorRight.tell(new SubPrintMessage(getContext().getSelf(),sub.right(), false));
        }
        else if (expression instanceof Expression.Mul mul) {
            this.operationString = "*";
            ActorRef<Message> subActorLeft = getContext().spawnAnonymous(Actor.create());
            subActorLeft.tell(new SubPrintMessage(getContext().getSelf(),mul.left(), true));
            ActorRef<Message> subActorRight = getContext().spawnAnonymous(Actor.create());
            subActorRight.tell(new SubPrintMessage(getContext().getSelf(),mul.right(), false));
        } else if (expression instanceof Expression.Val val) {
            if(side == true){
                oberActor.tell(new Actor.LeftStringResponse(String.valueOf(val.inner())));
            }else{
                oberActor.tell(new Actor.RightStringResponse(String.valueOf(val.inner())));
            }
        }
        return this;
    }

    //Antwort von dem linken Teilkind wird empfangen, verarbeitet und an den nächsthöheren Actor geschickt, falls von beiden Teilkindern
    //bereits eine Antwort erhalten wurde. Hier für den Formatierungsvorgang.
    private Behavior<Message> onLeftStringResponse(LeftStringResponse response){
        this.leftString = response.string;
        if(this.rightString != null){
           if(this.initial != null){
               initial.tell(new PrintAndEvaluate.StringResponse("(" + leftString + operationString + rightString +")"));
           }else {
               if(side == true){
                   oberActor.tell(new Actor.LeftStringResponse("(" + leftString + operationString + rightString +")"));
               }else {
                   oberActor.tell(new Actor.RightStringResponse("(" + leftString + operationString + rightString +")"));
               }
           }
        }
        return this;
    }
    //Antwort von dem rechten Teilkind wird empfangen, verarbeitet und an den nächsthöheren Actor geschickt, falls von beiden Teilkindern
    //bereits eine Antwort erhalten wurde. Hier für den Formatierungsvorgang.
    private Behavior<Message> onRightStringResponse(RightStringResponse response) {
        this.rightString = response.string;
        if (this.leftString != null) {
            if (this.initial != null) {
                initial.tell(new PrintAndEvaluate.StringResponse("(" + leftString + operationString + rightString + ")"));
            } else {
                if (side == true) {
                    oberActor.tell(new Actor.LeftStringResponse("(" + leftString + operationString + rightString + ")"));
                } else {
                    oberActor.tell(new Actor.RightStringResponse("(" + leftString + operationString + rightString + ")"));
                }
            }
        }
        return this;
    }

    //Die Nachricht von dem initialen Actor wird verarbeitet und, falls nötig (kein Val), weitere Actors erstellt.
    //Dabei wird seine Referenz in @initial gespeichert und diese ist für den Rechenvorgang
    private Behavior<Message> onEvaluateMessage(EvaluateMessage message){
        this.initial = message.sender;
        this.expression = message.expression;
        //Wenn initial nur Val übergeben wird → einfach ausgeben ansonsten UnterActors erstellen
        if (Objects.requireNonNull(expression) instanceof Expression.Add add) {
            this.operationString = "+";
            ActorRef<Actor.Message> LeftSubActor = getContext().spawnAnonymous(Actor.create());
            LeftSubActor.tell(new Actor.SubEvaluateMessage(getContext().getSelf(),add.left(), true));
            ActorRef<Actor.Message> RightSubActor = getContext().spawnAnonymous(Actor.create());
            RightSubActor.tell(new Actor.SubEvaluateMessage(getContext().getSelf(),add.right(), false));
        }
        else if (expression instanceof Expression.Sub sub) {
            this.operationString = "-";
            ActorRef<Actor.Message> LeftSubActor = getContext().spawnAnonymous(Actor.create());
            LeftSubActor.tell(new Actor.SubEvaluateMessage(getContext().getSelf(),sub.left(), true));
            ActorRef<Actor.Message> RightSubActor = getContext().spawnAnonymous(Actor.create());
            RightSubActor.tell(new Actor.SubEvaluateMessage(getContext().getSelf(),sub.right(), false));
        }
        else if (expression instanceof Expression.Mul mul) {
            this.operationString = "*";
            ActorRef<Actor.Message> LeftSubActor = getContext().spawnAnonymous(Actor.create());
            LeftSubActor.tell(new Actor.SubEvaluateMessage(getContext().getSelf(),mul.left(), true));
            ActorRef<Actor.Message> RightSubActor = getContext().spawnAnonymous(Actor.create());
            RightSubActor.tell(new Actor.SubEvaluateMessage(getContext().getSelf(),mul.right(), false));
        } else if (expression instanceof Expression.Val val) {
            initial.tell(new PrintAndEvaluate.IntResponse(val.inner()));
        }
        return this;
    }

    //Nachricht von dem Actor, von welchem dieser Actor erstellt wurde, wird verarbeitet und,
    //falls nötig (kein Val), weitere SubActors erstellt.
    //Dabei wird seine Referenz in oberActor gespeichert und diese Funktion ist für den Rechenvorgang
    private Behavior<Message> onSubEvaluateMessage(SubEvaluateMessage message){
        this.oberActor = message.sender;
        this.expression = message.expression;
        this.side = message.side;
        //Wenn initial nur Val übergeben wird → einfach ausgeben ansonsten UnterActors erstellen
        if (Objects.requireNonNull(expression) instanceof Expression.Add add) {
            this.operationString = "+";
            ActorRef<Actor.Message> LeftSubActor = getContext().spawnAnonymous(Actor.create());
            LeftSubActor.tell(new Actor.SubEvaluateMessage(getContext().getSelf(),add.left(), true));
            ActorRef<Actor.Message> RightSubActor = getContext().spawnAnonymous(Actor.create());
            RightSubActor.tell(new Actor.SubEvaluateMessage(getContext().getSelf(),add.right(), false));
        }
        else if (expression instanceof Expression.Sub sub) {
            this.operationString = "-";
            ActorRef<Actor.Message> LeftSubActor = getContext().spawnAnonymous(Actor.create());
            LeftSubActor.tell(new Actor.SubEvaluateMessage(getContext().getSelf(),sub.left(), true));
            ActorRef<Actor.Message> RightSubActor = getContext().spawnAnonymous(Actor.create());
            RightSubActor.tell(new Actor.SubEvaluateMessage(getContext().getSelf(),sub.right(), false));
        }
        else if (expression instanceof Expression.Mul mul) {
            this.operationString = "*";
            ActorRef<Actor.Message> LeftSubActor = getContext().spawnAnonymous(Actor.create());
            LeftSubActor.tell(new Actor.SubEvaluateMessage(getContext().getSelf(),mul.left(), true));
            ActorRef<Actor.Message> RightSubActor = getContext().spawnAnonymous(Actor.create());
            RightSubActor.tell(new Actor.SubEvaluateMessage(getContext().getSelf(),mul.right(), false));
        } else if (expression instanceof Expression.Val val) {
            if(side == true){
                oberActor.tell(new LeftIntResponse(val.inner()));
            }else{
                oberActor.tell(new RightIntResponse(val.inner()));
            }

        }
        return this;
    }

    //Antwort von dem linken Teilkind wird empfangen. Falls von beiden Teilkindern bereits eine Antwort erhalten wurde,
    //wird eine Sekunde lang gewartet und die Lösung des (Teil-)Ausdrucks berechnet.
    private Behavior<Message> onLeftIntResponse(LeftIntResponse response){
        this.leftInt = response.wert;
        if(this.rightInt != 0){
            timer.startSingleTimer(new Compute(), Duration.ofSeconds(1));
        }
        return this;
    }

    //Antwort von dem rechten Teilkind wird empfangen. Falls von beiden Teilkindern bereits eine Antwort erhalten wurde,
    //wird eine Sekunde lang gewartet und die Lösung des (Teil-)Ausdrucks berechnet.
    private Behavior<Message> onRightIntResponse(RightIntResponse response){
        this.rightInt = response.wert;
        if(this.leftInt != 0){
            timer.startSingleTimer(new Compute(), Duration.ofSeconds(1));
        }
        return this;
    }

    //Hier wird die Lösung des (Teil-)Ausdrucks berechnet und an den nächsthöheren Actor oder, falls nicht vorhanden,
    //an PrintAndEvaluate gesendet.
    private Behavior<Message> onComputeMessage(Compute response) {
        if (oberActor != null) {
            //side == true bedeutet, dass dieser Actor das linke Kind ist und es wird LeftResponse geschickt
            if (side == true) {
                if (operationString.equals("+")) {
                    oberActor.tell(new LeftIntResponse(leftInt + rightInt));
                } else if (operationString.equals("-")) {
                    oberActor.tell(new LeftIntResponse(leftInt - rightInt));
                } else {
                    oberActor.tell(new LeftIntResponse(leftInt * rightInt));
                }
            } else {
                if (operationString.equals("+")) {
                    oberActor.tell(new RightIntResponse(leftInt + rightInt));
                } else if (operationString.equals("-")) {
                    oberActor.tell(new RightIntResponse(leftInt - rightInt));
                } else {
                    oberActor.tell(new RightIntResponse(leftInt * rightInt));
                }
            }
        } else {
                if (operationString.equals("+")) {
                    initial.tell(new PrintAndEvaluate.IntResponse(leftInt + rightInt));
                } else if (operationString.equals("-")) {
                    initial.tell(new PrintAndEvaluate.IntResponse(leftInt - rightInt));
                } else {
                    initial.tell(new PrintAndEvaluate.IntResponse(leftInt * rightInt));
                }
        }
        return this;
    }
}
